from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Dict, List, Literal, Mapping, Optional, Union

from aibs_informatics_core.env import ENV_BASE_KEY_ALIAS, EnvBase, get_env_base
from aibs_informatics_core.models.aws.batch import JobName, ResourceRequirements
from aibs_informatics_core.utils.decorators import retry
from aibs_informatics_core.utils.hashing import sha256_hexdigest
from aibs_informatics_core.utils.logging import get_logger
from aibs_informatics_core.utils.tools.dicttools import convert_key_case
from aibs_informatics_core.utils.tools.strtools import pascalcase
from botocore.exceptions import ClientError

from aibs_informatics_aws_utils.core import AWS_REGION_VAR, AWSService, get_region
from aibs_informatics_aws_utils.logs import build_log_stream_url

if TYPE_CHECKING:  # pragma: no cover
    from mypy_boto3_batch.literals import JobDefinitionTypeType
    from mypy_boto3_batch.type_defs import (
        ContainerOverridesTypeDef,
        ContainerPropertiesTypeDef,
        DescribeJobsResponseTypeDef,
        EFSVolumeConfigurationTypeDef,
        EvaluateOnExitTypeDef,
        HostTypeDef,
        JobDefinitionTypeDef,
        KeyValuePairTypeDef,
        LinuxParametersTypeDef,
        MountPointTypeDef,
        RegisterJobDefinitionRequestTypeDef,
        RegisterJobDefinitionResponseTypeDef,
        ResourceRequirementTypeDef,
        RetryStrategyTypeDef,
        SubmitJobRequestTypeDef,
        SubmitJobResponseTypeDef,
        VolumeTypeDef,
    )
else:
    JobDefinitionTypeType = str
    RegisterJobDefinitionRequestTypeDef = dict
    ContainerOverridesTypeDef = dict
    ContainerPropertiesTypeDef = dict
    EFSVolumeConfigurationTypeDef = dict
    EvaluateOnExitTypeDef = dict
    HostTypeDef = dict
    JobDefinitionTypeDef = dict
    DescribeJobsResponseTypeDef = dict
    KeyValuePairTypeDef = dict
    LinuxParametersTypeDef = dict
    MountPointTypeDef = dict
    RegisterJobDefinitionResponseTypeDef = dict
    ResourceRequirementTypeDef = dict
    RetryStrategyTypeDef = dict
    SubmitJobRequestTypeDef = dict
    SubmitJobResponseTypeDef = dict
    VolumeTypeDef = dict


logger = get_logger(__name__)


get_batch_client = AWSService.BATCH.get_client


def to_volume(
    source_path: Optional[str],
    name: Optional[str],
    efs_volume_configuration: Optional[EFSVolumeConfigurationTypeDef],
) -> VolumeTypeDef:
    volume_dict = VolumeTypeDef()
    if source_path:
        volume_dict["host"] = HostTypeDef(sourcePath=source_path)
    if name:
        volume_dict["name"] = name
    if efs_volume_configuration:
        volume_dict["efsVolumeConfiguration"] = efs_volume_configuration
    return volume_dict


def to_mount_point(
    container_path: Optional[str],
    read_only: bool,
    source_volume: Optional[str],
) -> MountPointTypeDef:
    mount_point_dict = MountPointTypeDef(readOnly=read_only)
    if container_path:
        mount_point_dict["containerPath"] = container_path
    if source_volume:
        mount_point_dict["sourceVolume"] = source_volume
    return mount_point_dict


def to_key_value_pairs(
    environment: Dict[str, str],
    remove_null_values: bool = True,
) -> List[KeyValuePairTypeDef]:
    """Converts a map style of environment variables into a list of key-value pairs

    Args:
        environment (Dict[str, str]): map of environment variable keys and values
        remove_null_values (bool): Whether to withhold pairs where value is None. Defaults to True

    Returns:
        List[KeyValuePairTypeDef]: List of name,value json blobs representing env variables
    """

    return sorted(
        [
            KeyValuePairTypeDef(name=k, value=v)
            for k, v in environment.items()
            if not remove_null_values or v is not None
        ],
        key=lambda _: _.get("name", ""),
    )


def to_resource_requirements(
    gpu: Optional[int] = None,
    memory: Optional[int] = None,
    vcpus: Optional[int] = None,
) -> List[ResourceRequirementTypeDef]:
    """Converts Batch resource requirement parameters into a list of ResourceRequirement objects

    The returned list only includes dictionary entries for resources that specify
    an explicit value. Anything unset will be dropped.

    Args:
        gpu (Optional[int], optional): number of GPUs to use. Defaults to None.
        memory (Optional[int], optional): amount of memory in MiB. Defaults to None.
        vcpus (Optional[int], optional): Number of VCPUs to use. Defaults to None.

    Returns:
        List[ResourceRequirementTypeDef]: list of resource requirements
    """

    pairs: list[tuple[Literal["GPU", "MEMORY", "VCPU"], Optional[int]]] = [
        ("GPU", gpu),
        ("MEMORY", memory),
        ("VCPU", vcpus),
    ]
    return [ResourceRequirementTypeDef(type=t, value=str(v)) for t, v in pairs if v is not None]


def build_retry_strategy(
    num_retries: int = 5,
    evaluate_on_exit_configs: Optional[List[EvaluateOnExitTypeDef]] = None,
    include_default_evaluate_on_exit_configs: bool = True,
) -> RetryStrategyTypeDef:
    """Build a Retry Strategy for a Job definition

    By default, SPOT Termination retries are included. These can be excluded if desired.

    https://aws.amazon.com/blogs/compute/introducing-retry-strategies-for-aws-batch/

    Args:
        num_retries (int, optional): number of times to retry. Defaults to 5.
        evaluate_on_exit_list (list, optional): list of EvaluateOnExit configs.
        include_default_evaluate_on_exit_configs (bool, optional): Whether to exclude default
            evaluate on exit configuraitons
    Returns:
        RetryStrategy
    """
    all_evaluate_on_exit_configs: List[EvaluateOnExitTypeDef] = []
    if evaluate_on_exit_configs:
        all_evaluate_on_exit_configs.extend(evaluate_on_exit_configs)
    if include_default_evaluate_on_exit_configs:
        all_evaluate_on_exit_configs.extend(
            [
                EvaluateOnExitTypeDef(
                    action="RETRY",
                    onStatusReason="Task failed to start",
                    onReason="DockerTimeoutError*",
                ),
                EvaluateOnExitTypeDef(action="RETRY", onStatusReason="Host EC2*"),
                EvaluateOnExitTypeDef(action="EXIT", onStatusReason="*"),
            ]
        )
    return RetryStrategyTypeDef(attempts=num_retries, evaluateOnExit=all_evaluate_on_exit_configs)


# TODO: need better way of checking when parallel register job def calls collide
@retry(ClientError)
def register_job_definition(
    job_definition_name: str,
    container_properties: ContainerPropertiesTypeDef,
    parameters: Optional[Mapping[str, str]] = None,
    job_definition_type: JobDefinitionTypeType = "container",
    retry_strategy: Optional[RetryStrategyTypeDef] = None,
    tags: Optional[Mapping[str, str]] = None,
    propagate_tags: bool = False,
    region: Optional[str] = None,
) -> Union[JobDefinitionTypeDef, RegisterJobDefinitionResponseTypeDef]:
    batch = get_batch_client(region=region)

    # First we check to make sure that we aren't crearting unnecessary revisions
    # of the same job definition.
    latest = get_latest_job_definition(job_definition_name=job_definition_name, region=region)
    logger.info(f"Previously registered batch job definition: {latest}")
    if latest:
        latest_container_properties = latest.get("containerProperties", {})
        if (
            latest_container_properties.get("command") == container_properties.get("command")
            and latest_container_properties.get("image") == container_properties.get("image")
            and latest_container_properties.get("jobRoleArn")
            == container_properties.get("jobRoleArn")
            and latest.get("parameters") == parameters
            and latest.get("type") == job_definition_type
            and latest.get("tags") == tags
            and latest.get("retryStrategy") == retry_strategy
        ):
            logger.info(
                f"Latest job definition (name={job_definition_name}) matches expected. "
                "Skipping register new job definition call"
            )
            return latest
    register_job_definition_kwargs = RegisterJobDefinitionRequestTypeDef(
        jobDefinitionName=job_definition_name,
        type=job_definition_type,
        parameters=parameters or {},
        containerProperties=container_properties,
        propagateTags=propagate_tags,
        retryStrategy=retry_strategy or {},
        tags=tags or {},
    )
    logger.info(
        f"Registering job definition with following properties: {register_job_definition_kwargs}"
    )
    response = batch.register_job_definition(**register_job_definition_kwargs)  # type: ignore[arg-type]
    return response


def get_latest_job_definition(
    job_definition_name: str, region: Optional[str] = None
) -> Optional[JobDefinitionTypeDef]:
    batch = get_batch_client(region=region)
    response = batch.describe_job_definitions(
        jobDefinitionName=job_definition_name,
        maxResults=1,
        status="ACTIVE",
    )
    if len(response["jobDefinitions"]) == 1:
        return response["jobDefinitions"][0]
    elif len(response["jobDefinitions"]) > 1:
        return sorted(response["jobDefinitions"], key=lambda _: _["revision"], reverse=True)[0]
    return None


def submit_job(
    job_definition: str,
    job_queue: str,
    job_name: Optional[Union[JobName, str]] = None,
    env_base: Optional[EnvBase] = None,
    region: Optional[str] = None,
) -> SubmitJobResponseTypeDef:
    batch_client = get_batch_client(region=region)
    env_base = env_base or get_env_base()
    if job_name is None:
        job_name = JobName(f"{env_base}-{sha256_hexdigest()}")
    else:
        job_name = JobName(job_name)
    submit_job_kwargs = SubmitJobRequestTypeDef(
        jobName=job_name,
        jobQueue=job_queue,
        jobDefinition=job_definition,
    )
    response = batch_client.submit_job(**submit_job_kwargs)
    return response


@dataclass
class BatchJobBuilder:
    image: str
    job_definition_name: str
    job_name: str
    command: List[str] = field(default_factory=list)
    environment: Dict[str, str] = field(default_factory=dict)
    job_definition_tags: Dict[str, str] = field(default_factory=dict)
    resource_requirements: Union[List[ResourceRequirementTypeDef], ResourceRequirements] = field(
        default_factory=list
    )
    mount_points: List[MountPointTypeDef] = field(default_factory=list)
    volumes: List[VolumeTypeDef] = field(default_factory=list)
    job_role_arn: Optional[str] = field(default=None)
    privileged: bool = field(default=False)
    linux_parameters: Optional[LinuxParametersTypeDef] = field(default=None)
    env_base: EnvBase = field(default_factory=EnvBase.from_env)

    def __post_init__(self):
        self.environment[ENV_BASE_KEY_ALIAS] = self.env_base
        self.environment[AWS_REGION_VAR] = get_region()
        # TODO: Should add EFS mount point if present as env var

    @property
    def container_properties(self) -> ContainerPropertiesTypeDef:
        container_props = ContainerPropertiesTypeDef(
            image=self.image,
            command=self.command,
            environment=to_key_value_pairs(self.environment),
            resourceRequirements=self._normalized_resource_requirements(),
            mountPoints=self.mount_points,
            volumes=self.volumes,
            privileged=self.privileged,
        )
        if self.linux_parameters:
            container_props["linuxParameters"] = self.linux_parameters
        if self.job_role_arn:
            container_props["jobRoleArn"] = self.job_role_arn
        return container_props

    @property
    def container_overrides(self) -> ContainerOverridesTypeDef:
        return ContainerOverridesTypeDef(
            environment=to_key_value_pairs(self.environment),
            resourceRequirements=self._normalized_resource_requirements(),
        )

    @property
    def container_overrides__sfn(self) -> Dict[str, Any]:
        return convert_key_case(self.container_overrides, pascalcase)  # type: ignore[arg-type]

    def _normalized_resource_requirements(self) -> List[ResourceRequirementTypeDef]:
        if isinstance(self.resource_requirements, list):
            return sorted(self.resource_requirements, key=lambda _: _["type"])
        else:
            return to_resource_requirements(
                gpu=self.resource_requirements.GPU,
                memory=self.resource_requirements.MEMORY,
                vcpus=self.resource_requirements.VCPU,
            )


def describe_jobs(
    job_ids: List[str],
    region: Optional[str] = None,
) -> DescribeJobsResponseTypeDef:
    batch = get_batch_client(region=region)
    response = batch.describe_jobs(jobs=job_ids)
    return response


def batch_log_stream_name_to_url(log_stream_name: str, region: Optional[str] = None) -> str:
    log_group_name = "/aws/batch/job"
    return build_log_stream_url(
        log_group_name=log_group_name, log_stream_name=log_stream_name, region=region
    )



__all__ = [
    "DataSyncOperations",
    "LocalFileSystem",
    "Node",
    "S3FileSystem",
    "sync_data",
]

from .file_system import LocalFileSystem, Node, S3FileSystem
from .operations import DataSyncOperations, sync_data



from copy import deepcopy
from decimal import Decimal
from typing import TYPE_CHECKING, Any, Dict, List, Literal, Mapping, Optional, Union

from aibs_informatics_core.utils.logging import get_logger
from boto3.dynamodb.conditions import ConditionBase
from boto3.dynamodb.types import TypeDeserializer, TypeSerializer

from aibs_informatics_aws_utils.core import AWSService
from aibs_informatics_aws_utils.dynamodb.conditions import (
    ConditionExpressionComponents,
    UpdateExpressionComponents,
)

if TYPE_CHECKING:  # pragma: no cover
    from mypy_boto3_dynamodb.type_defs import (
        BatchGetItemInputTypeDef,
        GetItemInputTypeDef,
        KeysAndAttributesTypeDef,
        QueryInputTableQueryTypeDef,
        QueryInputTypeDef,
        ScanInputTypeDef,
    )
else:
    BatchGetItemInputTypeDef = dict
    GetItemInputRequestTypeDef = dict
    GetItemInputTypeDef = dict
    KeysAndAttributesTypeDef = dict
    QueryInputTableQueryTypeDef = dict
    QueryInputTypeDef = dict
    ScanInputTypeDef = dict


logger = get_logger(__name__)


get_dynamodb_client = AWSService.DYNAMO_DB.get_client
get_dynamodb_resource = AWSService.DYNAMO_DB.get_resource


# ----------------------------------------------------------------------------
# Dynamo DB Table Methods
# ----------------------------------------------------------------------------
def table_put_item(
    table_name: str,
    item: Dict[str, Any],
    condition_expression: Optional[ConditionBase] = None,
    **kwargs,
):
    if condition_expression:
        parsed_expression = ConditionExpressionComponents.from_condition(
            condition_expression, False
        )
        kwargs["ConditionExpression"] = parsed_expression.condition_expression
        kwargs["ExpressionAttributeNames"] = parsed_expression.expression_attribute_names
        # Not always necessary to have ExpressionAttributeValues
        if parsed_expression.expression_attribute_values:
            kwargs["ExpressionAttributeValues"] = parsed_expression.expression_attribute_values

    # For more details on additional kwargs for table.put_item see:
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/dynamodb.html#DynamoDB.Table.put_item
    table = table_as_resource(table_name)
    return table.put_item(Item=item, **kwargs)


def table_get_item(
    table_name: str, key: Mapping[str, Any], attrs: Optional[str] = None
) -> Optional[Dict[str, Any]]:
    table = table_as_resource(table_name)
    props: GetItemInputRequestTypeDef = {"Key": key, "ReturnConsumedCapacity": "NONE"}  # type: ignore  # we modify use of this type (no table name is needed here)

    if attrs is not None:
        props["ProjectionExpression"] = attrs

    response = table.get_item(**props)  # type: ignore  # pylance complains about extra fields

    logger.info("Response from table.get_item: %s", response)

    return response.get("Item", None)


def table_get_items(
    table_name: str,
    keys: List[Mapping[str, Any]],
    attrs: Optional[str] = None,
    region: Optional[str] = None,
) -> List[Dict[str, Any]]:
    db = get_dynamodb_client(region=region)
    serializer = TypeSerializer()

    items: List[Dict[str, Any]] = []

    # we receive an error if there are more than 100 calls
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/dynamodb.html#DynamoDB.Client.batch_get_item
    MAX_KEYS_PER_API_CALL = 100

    keys_subset_list = [
        keys[i : i + MAX_KEYS_PER_API_CALL] for i in range(0, len(keys), MAX_KEYS_PER_API_CALL)
    ]
    for keys_subset in keys_subset_list:
        serialized_keys = [
            {key: serializer.serialize(value) for key, value in attr_value.items()}
            for attr_value in keys_subset
        ]

        request_items: Mapping[str, KeysAndAttributesTypeDef] = {
            table_name: {
                "Keys": serialized_keys,
            }
        }
        props: BatchGetItemInputTypeDef = {
            "RequestItems": request_items,
            "ReturnConsumedCapacity": "NONE",
        }

        if attrs is not None:
            for _, keys_and_attrs in request_items.items():
                keys_and_attrs["ProjectionExpression"] = attrs

        while True:
            response = db.batch_get_item(**props)
            for table_items in response["Responses"].values():
                items.extend(table_items)
            # Must make subsequent calls for Unprocessed keys if present.
            # https://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_BatchGetItem.html#DDB-BatchGetItem-response-UnprocessedKeys
            if response.get("UnprocessedKeys", None):
                props["RequestItems"] = response["UnprocessedKeys"]
            else:
                # If no more keys to process, break from while loop.
                break

    deserializer = TypeDeserializer()
    return [{k: deserializer.deserialize(v) for k, v in item.items()} for item in items]


def table_update_item(
    table_name: str,
    key: Mapping[str, Any],
    attributes: Mapping[str, Any],
    return_values: Literal["NONE", "ALL_OLD", "UPDATED_OLD", "ALL_NEW", "UPDATED_NEW"] = "NONE",
    **kwargs,
) -> Optional[Dict[str, Any]]:
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/dynamodb/table/update_item.html
    table = table_as_resource(table_name)

    expression_components = UpdateExpressionComponents.from_dict(attributes=attributes)

    response = table.update_item(
        Key=key,
        ExpressionAttributeNames=expression_components.expression_attribute_names,
        ExpressionAttributeValues=expression_components.expression_attribute_values,
        UpdateExpression=expression_components.update_expression,
        ReturnValues=return_values,
        **kwargs,
    )

    logger.info(f"Response from table.update_item: {response}")

    return response.get("Attributes", None)


def table_delete_item(
    table_name: str,
    key: Mapping[str, Any],
    condition_expression: Optional[ConditionBase] = None,
    return_values: Literal["NONE", "ALL_OLD"] = "NONE",
    **kwargs,
) -> Optional[Dict[str, Any]]:
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/dynamodb/table/delete_item.html

    if condition_expression:
        parsed_expression = ConditionExpressionComponents.from_condition(
            condition_expression, False
        )
        kwargs["ConditionExpression"] = parsed_expression.condition_expression
        kwargs["ExpressionAttributeNames"] = parsed_expression.expression_attribute_names
        kwargs["ExpressionAttributeValues"] = parsed_expression.expression_attribute_values

    table = table_as_resource(table_name)
    response = table.delete_item(Key=key, ReturnValues=return_values, **kwargs)
    return response.get("Attributes", None)


def table_query(
    table_name: str,
    key_condition_expression: ConditionBase,
    index_name: Optional[str] = None,
    filter_expression: Optional[ConditionBase] = None,
    region: Optional[str] = None,
    consistent_read: bool = False,
) -> List[Dict[str, Any]]:
    """Query a table

    Args:
        table_name (str): Name of the table (prefixed with EnvBase)
        index_name (str, optional): Index name. Defaults to None.
        key_condition_expression (ConditionBase): Key condition expression.
        filter_expression (ConditionBase, optional): Filter expression. Defaults to None.
        env_base (EnvBase, optional): EnvBase. Defaults to None.
        region (str, optional): Region. Defaults to None.
        consistent_read (bool, optional): Whether a strongly consistent read should be used for the
            query. Defaults to False.
            NOTE: Strongly consistent reads are not supported for global secondary indexes

    Returns:
        List[Dict[str, Any]]: Items queried
    """
    db = get_dynamodb_client(region=region)
    table = table_as_resource(table_name)

    key_expr_component = ConditionExpressionComponents.from_condition(
        key_condition_expression, True
    )
    expression_attribute_names = deepcopy(key_expr_component.expression_attribute_names)
    expression_attribute_values = deepcopy(
        key_expr_component.expression_attribute_values__serialized
    )

    db_request: QueryInputTypeDef = {
        "TableName": table.name,
        "KeyConditionExpression": key_expr_component.condition_expression,
    }

    # Handle when filter_expression is provided
    if filter_expression is not None:
        filter_expr_component = ConditionExpressionComponents.from_condition(
            filter_expression, False
        )
        # For queries, there is a possibility that key/sort and filter expression
        # attribute names could collide
        clean_filter_expr_components = key_expr_component.fix_collisions(filter_expr_component)
        expression_attribute_names.update(clean_filter_expr_components.expression_attribute_names)
        expression_attribute_values.update(
            clean_filter_expr_components.expression_attribute_values__serialized
        )
        db_request["FilterExpression"] = clean_filter_expr_components.condition_expression

    # ExpressionAttributeNames/Values should always exist for `query`
    db_request["ExpressionAttributeNames"] = expression_attribute_names
    if expression_attribute_values:
        db_request["ExpressionAttributeValues"] = expression_attribute_values

    # Handle a query/scan involving an index (GSI/LSI)
    if index_name is not None:
        db_request["IndexName"] = index_name
    if index_name is not None and consistent_read is True:
        logger.warning(
            f"Strongly consistent reads are NOT supported for secondary indices "
            f"like '{index_name}' for the table '{table_name}'. Ignoring provided "
            "`consistent_read` value and treating it as False!"
        )
    else:
        db_request["ConsistentRead"] = consistent_read

    items: List[Dict[str, Any]] = []
    paginator = db.get_paginator("query")
    logger.info(f"Performing DB 'query' on {table.name} with following parameters: {db_request}")
    for i, response in enumerate(paginator.paginate(**db_request)):  # type: ignore  # pylance complains about extra fields
        new_items = response.get("Items", [])
        items.extend(new_items)
        logger.debug(f"Iter #{i + 1}: item count from table. Query: {len(new_items)}")
    logger.info(f"Complete item count from table. Query (after any filtering): {len(items)}")

    deserializer = TypeDeserializer()
    return [{k: deserializer.deserialize(v) for k, v in item.items()} for item in items]


def table_scan(
    table_name: str,
    index_name: Optional[str] = None,
    filter_expression: Optional[ConditionBase] = None,
    region: Optional[str] = None,
    consistent_read: bool = False,
) -> List[Dict[str, Any]]:
    """Scan a table

    Args:
        table_name (str): Name of the table (prefixed with EnvBase)
        index_name (str, optional): Index name. Defaults to None.
        filter_expression (ConditionBase, optional): Filter expression. Defaults to None.
        region (str, optional): Region. Defaults to None.
        consistent_read (bool, optional): Whether a strongly consistent read should be used for the
            query. Defaults to False.
            NOTE: Strongly consistent reads are not supported for secondary indexes

    Returns:
        List[Dict[str, Any]]: Items queried
    """

    db = get_dynamodb_client(region=region)
    table = table_as_resource(table_name)

    db_request: ScanInputTypeDef = {"TableName": table.name}

    # Handle when filter_expression is provided
    if filter_expression is not None:
        filter_expr_component = ConditionExpressionComponents.from_condition(
            filter_expression, False
        )
        expression_attribute_names = deepcopy(filter_expr_component.expression_attribute_names)
        expression_attribute_values = deepcopy(
            filter_expr_component.expression_attribute_values__serialized
        )

        db_request["FilterExpression"] = filter_expr_component.condition_expression
        db_request["ExpressionAttributeNames"] = expression_attribute_names
        if expression_attribute_values:
            db_request["ExpressionAttributeValues"] = expression_attribute_values

    # Handle a scan involving an index (GSI/LSI)
    if index_name is not None:
        db_request["IndexName"] = index_name
    if index_name is not None and consistent_read is True:
        logger.warning(
            f"Strongly consistent reads are NOT supported for secondary indices "
            f"like '{index_name}' for the table '{table_name}'. Ignoring provided"
            " `consistent_read` value and treating it as False!"
        )
    else:
        db_request["ConsistentRead"] = consistent_read

    items: List[Dict[str, Any]] = []
    paginator = db.get_paginator("scan")
    logger.info(f"Performing DB 'scan' on {table.name} with following parameters: {db_request}")
    for i, response in enumerate(paginator.paginate(**db_request)):  # type: ignore  # pylance complains about extra fields
        new_items = response.get("Items", [])
        items.extend(new_items)
        logger.debug(f"Iter #{i + 1}: item count from table. Scan: {len(new_items)}")
    logger.info(
        f"Complete item count from table. Scan results (after any filtering): {len(items)}"
    )

    deserializer = TypeDeserializer()
    return [{k: deserializer.deserialize(v) for k, v in item.items()} for item in items]


def table_get_key_schema(table_name: str) -> Dict[str, str]:
    """Helper function to return mapping of key type to attribute name for a table
    e.g. {"HASH": PartitionKeyAttributeName, "RANGE": SortKeyAttributeName}
    """
    table = table_as_resource(table_name)
    return {k["KeyType"]: k["AttributeName"] for k in table.key_schema}


def execute_partiql_statement(
    statement: str, region: Optional[str] = None
) -> List[Dict[str, Any]]:
    db = get_dynamodb_client(region=region)

    response = db.execute_statement(Statement=statement)
    results = response["Items"]

    while response.get("NextToken", None):
        response = db.execute_statement(Statement=statement, NextToken=response["NextToken"])
        results.extend(response["Items"])
    return results


def table_as_resource(table: str, region: Optional[str] = None):
    """Helper method to get the table as a resource for given env_label
    if provided.
    """
    db = get_dynamodb_resource(region=region)
    return db.Table(table)


def convert_decimals_to_floats(item: Dict[str, Any], in_place: bool = True) -> Dict[str, Any]:
    def _convert_decimals_to_floats(obj: Union[Dict[str, Any], List[Any]]):
        if isinstance(obj, list):
            for i in range(len(obj)):
                if isinstance(obj[i], Decimal):
                    obj[i] = float(obj[i])
                elif isinstance(obj[i], (dict, list)):
                    _convert_decimals_to_floats(obj[i])
        elif isinstance(obj, dict):
            for k in obj:
                if isinstance(obj[k], Decimal):
                    obj[k] = float(obj[k])
                elif isinstance(obj[k], (dict, list)):
                    _convert_decimals_to_floats(obj[k])

    if not in_place:
        item = deepcopy(item)
    _convert_decimals_to_floats(item)
    return item


def convert_floats_to_decimals(item: Dict[str, Any], in_place: bool = True) -> Dict[str, Any]:
    def _convert_floats_to_decimals(obj: Union[Dict[str, Any], List[Any]]):
        if isinstance(obj, list):
            for i in range(len(obj)):
                if isinstance(obj[i], float):
                    obj[i] = Decimal(str(obj[i]))
                elif isinstance(obj[i], (dict, list)):
                    _convert_floats_to_decimals(obj[i])
        elif isinstance(obj, dict):
            for k in obj:
                if isinstance(obj[k], float):
                    obj[k] = Decimal(str(obj[k]))
                elif isinstance(obj[k], (dict, list)):
                    _convert_floats_to_decimals(obj[k])

    if not in_place:
        item = deepcopy(item)
    _convert_floats_to_decimals(item)
    return item



__all__ = [
    "AttrPlaceholder",
    "AttributeBaseExpression",
    "ConditionBaseExpression",
    "ConditionBaseExpressionString",
    "ConditionBaseTranslator",
    "convert_decimals_to_floats",
    "convert_floats_to_decimals",
    "execute_partiql_statement",
    "get_dynamodb_client",
    "get_dynamodb_resource",
    "table_as_resource",
    "table_delete_item",
    "table_get_item",
    "table_get_items",
    "table_get_key_schema",
    "table_put_item",
    "table_query",
    "table_scan",
    "table_update_item",
    "DynamoDBTable",
    "DynamoDBEnvBaseTable",
    "build_optimized_condition_expression_set",
    "check_db_query_non_empty",
    "check_db_query_unique",
    "check_index_supports_strongly_consistent_read",
    "check_table_name_and_index_match",
]

from .conditions import (
    AttributeBaseExpression,
    AttrPlaceholder,
    ConditionBaseExpression,
    ConditionBaseExpressionString,
    ConditionBaseTranslator,
)
from .functions import (
    convert_decimals_to_floats,
    convert_floats_to_decimals,
    execute_partiql_statement,
    get_dynamodb_client,
    get_dynamodb_resource,
    table_as_resource,
    table_delete_item,
    table_get_item,
    table_get_items,
    table_get_key_schema,
    table_put_item,
    table_query,
    table_scan,
    table_update_item,
)
from .table import (
    DynamoDBEnvBaseTable,
    DynamoDBTable,
    build_optimized_condition_expression_set,
    check_db_query_non_empty,
    check_db_query_unique,
    check_index_supports_strongly_consistent_read,
    check_table_name_and_index_match,
)



import logging
import sys
import time
from typing import TYPE_CHECKING, List, Literal, Optional, Tuple

if sys.version_info >= (3, 11):
    # For Python 3.11+
    from typing import Unpack
else:  # pragma: no cover
    # For Python < 3.11
    from typing_extensions import Unpack


from botocore.exceptions import ClientError

from aibs_informatics_aws_utils.core import AWSService
from aibs_informatics_aws_utils.exceptions import AWSError

if TYPE_CHECKING:  # pragma: no cover
    from mypy_boto3_athena.type_defs import (
        GetQueryExecutionOutputTypeDef,
        QueryExecutionStatusTypeDef,
        QueryExecutionTypeDef,
        StartQueryExecutionInputTypeDef,
        StartQueryExecutionOutputTypeDef,
    )
else:
    GetQueryExecutionOutputTypeDef = dict
    QueryExecutionStatusTypeDef = dict
    QueryExecutionTypeDef = dict
    StartQueryExecutionInputTypeDef = dict
    StartQueryExecutionOutputTypeDef = dict


ATHENA_QUERY_WAITER_STATUS = Literal["SUCCEEDED", "FAILED", "CANCELLED", "TIMEOUT"]

logger = logging.getLogger(__name__)

get_athena_client = AWSService.ATHENA.get_client


def start_query_execution(
    query_string: str,
    work_group: Optional[str] = None,
    execution_parameters: Optional[List[str]] = None,
    **kwargs: Unpack[StartQueryExecutionInputTypeDef],
) -> StartQueryExecutionOutputTypeDef:
    athena = get_athena_client()

    request = StartQueryExecutionInputTypeDef(QueryString=query_string)
    if work_group:
        request["WorkGroup"] = work_group
    if execution_parameters:
        request["ExecutionParameters"] = execution_parameters
    request.update(kwargs)
    try:
        metadata = athena.start_query_execution(**request)
        return metadata
    except ClientError as e:
        logger.error(f"Error executing : {request} {e}", exc_info=True)
        raise AWSError(f"Error starting query execution: {request} {e}") from e


def get_query_execution(query_execution_id: str) -> GetQueryExecutionOutputTypeDef:
    athena = get_athena_client()
    try:
        return athena.get_query_execution(QueryExecutionId=query_execution_id)
    except Exception as e:
        logger.error(f"Error executing : {query_execution_id} {e}", exc_info=True)
        raise AWSError(f"Error starting query execution: {query_execution_id} {e}") from e


def query_waiter(
    query_execution_id: str, timeout: int = 60
) -> Tuple[ATHENA_QUERY_WAITER_STATUS, QueryExecutionStatusTypeDef]:
    start = time.time()
    logger.info(f"Polling for status of query execution: {query_execution_id}")
    while True:
        stats = get_query_execution(query_execution_id=query_execution_id)
        logger.info(f"Query Execution Status: {stats}")
        status = stats["QueryExecution"].get("Status", {})
        state = status.get("State")
        if state in ["SUCCEEDED", "FAILED", "CANCELLED", "TIMEOUT"]:
            return state, status  # type: ignore[return-value]
        time.sleep(0.2)  # 200ms
        # Exit if the time waiting exceed the timeout seconds
        if time.time() > start + timeout:
            return "TIMEOUT", status



from __future__ import annotations

__all__ = [
    "get_efs_client",
    "list_efs_file_systems",
    "get_efs_file_system",
    "list_efs_access_points",
    "get_efs_access_point",
]

import logging
from pathlib import Path
from typing import TYPE_CHECKING, Dict, List, Optional, Union

from aibs_informatics_core.utils.decorators import retry
from aibs_informatics_core.utils.tools.dicttools import remove_null_values
from botocore.exceptions import ClientError

from aibs_informatics_aws_utils.core import AWSService, client_error_code_check

if TYPE_CHECKING:  # pragma: no cover
    from mypy_boto3_efs.type_defs import (
        AccessPointDescriptionTypeDef,
        DescribeAccessPointsResponseTypeDef,
        DescribeFileSystemsResponseTypeDef,
        FileSystemDescriptionTypeDef,
        TagTypeDef,
    )
else:
    AccessPointDescriptionTypeDef = dict
    DescribeFileSystemsResponseTypeDef = dict
    DescribeAccessPointsResponseTypeDef = dict
    FileSystemDescriptionTypeDef = dict
    TagTypeDef = dict


logger = logging.getLogger(__name__)

get_efs_client = AWSService.EFS.get_client

StrPath = Union[Path, str]


def throttling_exception_callback(ex):
    return client_error_code_check(ex, "ThrottlingException")


@retry(ClientError, [throttling_exception_callback])
def list_efs_file_systems(
    file_system_id: Optional[str] = None,
    name: Optional[str] = None,
    tags: Optional[Dict[str, str]] = None,
) -> List[FileSystemDescriptionTypeDef]:
    """List EFS file systems.

    You can filter on id, name and tags.

    Args:
        file_system_id (Optional[str], optional): Optionally filter on file system id.
        name (Optional[str], optional): Optionally filter on name.
        tags (Optional[Dict[str, str]], optional): Optionally filter on tags.
            They should be a dict of key-value pairs.

    Returns:
        List[FileSystemDescriptionTypeDef]: List of matching file systems
    """
    efs = get_efs_client()
    paginator = efs.get_paginator("describe_file_systems")

    file_systems: List[FileSystemDescriptionTypeDef] = []
    paginator_kwargs = remove_null_values(dict(FileSystemId=file_system_id))
    for results in paginator.paginate(**paginator_kwargs):  # type: ignore
        for fs in results["FileSystems"]:
            if name and fs.get("Name") != name:
                continue
            if tags:
                fs_tags = {tag["Key"]: tag["Value"] for tag in fs["Tags"]}
                if not all([tags[k] == fs_tags.get(k) for k in tags]):
                    continue
            file_systems.append(fs)
    return file_systems


def get_efs_file_system(
    file_system_id: Optional[str] = None,
    name: Optional[str] = None,
    tags: Optional[Dict[str, str]] = None,
) -> FileSystemDescriptionTypeDef:
    """Get EFS file system.

    You can filter on id, name and tags.

    Args:
        file_system_id (Optional[str], optional): Optionally filter on file system id.
        name (Optional[str], optional): Optionally filter on name.
        tags (Optional[Dict[str, str]], optional): Optionally filter on tags.
            They should be a dict of key-value pairs.

    Raises:
        ValueError: If no file system is found based on the filters.
        ValueError: If more than one file system is found based on the filters.

    Returns:
        FileSystemDescriptionTypeDef: The file system description.
    """
    file_systems = list_efs_file_systems(file_system_id=file_system_id, name=name, tags=tags)
    if len(file_systems) > 1:
        raise ValueError(
            f"Found more than one file systems ({len(file_systems)}) "
            f"based on id={file_system_id}, name={name}, tags={tags}"
        )
    elif len(file_systems) == 0:
        raise ValueError(
            f"Found no file systems based on id={file_system_id}, name={name}, tags={tags}"
        )
    return file_systems[0]


@retry(ClientError, [throttling_exception_callback])
def list_efs_access_points(
    access_point_id: Optional[str] = None,
    access_point_name: Optional[str] = None,
    access_point_tags: Optional[Dict[str, str]] = None,
    file_system_id: Optional[str] = None,
    file_system_name: Optional[str] = None,
    file_system_tags: Optional[Dict[str, str]] = None,
) -> List[AccessPointDescriptionTypeDef]:
    """List EFS access points.

    You can filter on id, name and tags for both access point and file system.

    Args:
        access_point_id (Optional[str], optional): Optionally filter on access point id.
        access_point_name (Optional[str], optional): Optionally filter on name.
        access_point_tags (Optional[Dict[str, str]], optional): Optionally filter on access point
            tags. They should be a dict of key-value pairs.
        file_system_id (Optional[str], optional): Optionally filter on file system id.
        file_system_name (Optional[str], optional): Optionally filter on file system name.
        file_system_tags (Optional[Dict[str, str]], optional): Optionally filter on file system
            tags. They should be a dict of key-value pairs.

    Returns:
        List[AccessPointDescriptionTypeDef]: List of matching access points
    """
    efs = get_efs_client()

    file_system_ids: List[str] = []
    if file_system_id:
        file_system_ids.append(file_system_id)
    elif file_system_name or file_system_tags:
        file_systems = list_efs_file_systems(
            file_system_id=file_system_id, name=file_system_name, tags=file_system_tags
        )
        file_system_ids.extend(map(lambda _: _["FileSystemId"], file_systems))

    access_points: List[AccessPointDescriptionTypeDef] = []

    if access_point_id or not file_system_ids:
        response = efs.describe_access_points(
            **remove_null_values(dict(AccessPointId=access_point_id))  # type: ignore
        )
        # If file_system_ids is empty, we want to include all access points. Otherwise,
        # we only want to include access points that belong to the file systems
        # in file_system_ids.
        for access_point in response["AccessPoints"]:
            if not file_system_ids or access_point.get("FileSystemId") in file_system_ids:
                access_points.append(access_point)
    else:
        for fs_id in file_system_ids:
            response = efs.describe_access_points(FileSystemId=fs_id)
            access_points.extend(response["AccessPoints"])
            while next_token := response.get("NextToken"):
                response = efs.describe_access_points(FileSystemId=fs_id, NextToken=next_token)
                access_points.extend(response["AccessPoints"])

    filtered_access_points: List[AccessPointDescriptionTypeDef] = []

    for ap in access_points:
        if access_point_name and ap.get("Name") != access_point_name:
            continue
        if access_point_tags:
            ap_tags = {tag["Key"]: tag["Value"] for tag in ap.get("Tags", {})}
            tags_match = [access_point_tags[k] == ap_tags.get(k) for k in access_point_tags]
            if not all(tags_match):
                continue
        filtered_access_points.append(ap)
    return filtered_access_points


def get_efs_access_point(
    access_point_id: Optional[str] = None,
    access_point_name: Optional[str] = None,
    access_point_tags: Optional[Dict[str, str]] = None,
    file_system_id: Optional[str] = None,
    file_system_name: Optional[str] = None,
    file_system_tags: Optional[Dict[str, str]] = None,
) -> AccessPointDescriptionTypeDef:
    """Get EFS access point.

    You can filter on id, name and tags for both access point and file system.

    Args:
        access_point_id (Optional[str], optional): Optionally filter on access point id.
        access_point_name (Optional[str], optional): Optionally filter on name.
        access_point_tags (Optional[Dict[str, str]], optional): Optionally filter on access point
            tags. They should be a dict of key-value pairs.
        file_system_id (Optional[str], optional): Optionally filter on file system id.
        file_system_name (Optional[str], optional): Optionally filter on file system name.
        file_system_tags (Optional[Dict[str, str]], optional): Optionally filter on file system
            tags. They should be a dict of key-value pairs.

    Raises:
        ValueError: If no access point is found based on the filters.
        ValueError: If more than one access point is found based on the filters.

    Returns:
        AccessPointDescriptionTypeDef: The access point description.
    """
    access_points = list_efs_access_points(
        access_point_id=access_point_id,
        access_point_name=access_point_name,
        access_point_tags=access_point_tags,
        file_system_id=file_system_id,
        file_system_name=file_system_name,
        file_system_tags=file_system_tags,
    )
    if len(access_points) > 1:
        raise ValueError(
            f"Found more than one access points ({len(access_points)}) "
            f"based on access point filters (id={access_point_id}, "
            f"name={access_point_name}, tags={access_point_tags}) "
            f"and on file system filters (id={file_system_id}, "
            f"name={file_system_name}, tags={file_system_tags}) "
        )
    elif len(access_points) == 0:
        raise ValueError(
            f"Found no access points "
            f"based on access point filters (id={access_point_id}, "
            f"name={access_point_name}, tags={access_point_tags}) "
            f"and on file system filters (id={file_system_id}, "
            f"name={file_system_name}, tags={file_system_tags}) "
        )
    return access_points[0]



import functools
import os
import tempfile
from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path
from typing import Optional, Union, cast

from aibs_informatics_core.models.aws.efs import EFSPath
from aibs_informatics_core.models.aws.s3 import S3URI, S3KeyPrefix
from aibs_informatics_core.models.data_sync import (
    DataSyncConfig,
    DataSyncRequest,
    DataSyncResult,
    DataSyncTask,
    RemoteToLocalConfig,
)
from aibs_informatics_core.utils.decorators import retry
from aibs_informatics_core.utils.file_operations import (
    CannotAcquirePathLockError,
    PathLock,
    copy_path,
    find_filesystem_boundary,
    get_path_size_bytes,
    move_path,
    remove_path,
)
from aibs_informatics_core.utils.logging import LoggingMixin, get_logger
from aibs_informatics_core.utils.os_operations import find_all_paths
from botocore.client import Config

from aibs_informatics_aws_utils.efs import get_local_path
from aibs_informatics_aws_utils.s3 import (
    TransferConfig,
    delete_s3_path,
    get_s3_path_stats,
    is_folder,
    is_object,
    sync_paths,
)

logger = get_logger(__name__)


MAX_LOCK_WAIT_TIME_IN_SECS = 60 * 60 * 6  # 6 hours

LOCK_ROOT_ENV_VAR = "DATA_SYNC_LOCK_ROOT"

LocalPath = Union[Path, EFSPath]


@functools.cache
def get_botocore_config(max_pool_connections: int, **kwargs) -> Config:
    return Config(max_pool_connections=max_pool_connections, **kwargs)


@dataclass
class DataSyncOperations(LoggingMixin):
    config: DataSyncConfig

    @property
    def s3_transfer_config(self) -> TransferConfig:
        return TransferConfig(max_concurrency=self.config.max_concurrency)

    @property
    def botocore_config(self) -> Config:
        return get_botocore_config(max_pool_connections=self.config.max_concurrency)

    def sync_local_to_s3(self, source_path: LocalPath, destination_path: S3URI) -> DataSyncResult:
        source_path = self.sanitize_local_path(source_path)
        if not source_path.exists():
            if self.config.fail_if_missing:
                raise FileNotFoundError(f"Local path {source_path} does not exist")
            self.logger.warning(f"Local path {source_path} does not exist")
            if self.config.include_detailed_response:
                return DataSyncResult(bytes_transferred=0, files_transferred=0)
            else:
                return DataSyncResult()
        if source_path.is_dir():
            self.logger.info("local source path is folder. Adding suffix to destination path")
            destination_path = S3URI.build(
                bucket_name=destination_path.bucket_name,
                key=destination_path.key_with_folder_suffix,
            )
        self.logger.info(f"Uploading local content from {source_path} -> {destination_path}")
        sync_paths(
            source_path=source_path,
            destination_path=destination_path,
            transfer_config=self.s3_transfer_config,
            config=self.botocore_config,
            force=self.config.force,
            size_only=self.config.size_only,
            delete=True,
        )
        result = DataSyncResult()
        if self.config.include_detailed_response:
            result.files_transferred = len(find_all_paths(source_path, include_dirs=False))
            result.bytes_transferred = get_path_size_bytes(source_path)
        if not self.config.retain_source_data:
            remove_path(source_path)
        return result

    def sync_s3_to_local(self, source_path: S3URI, destination_path: LocalPath) -> DataSyncResult:
        self.logger.info(f"Downloading s3 content from {source_path} -> {destination_path}")
        start_time = datetime.now(tz=timezone.utc)
        destination_path = self.sanitize_local_path(destination_path)
        source_is_object = is_object(source_path)

        if not source_is_object and not is_folder(source_path):
            message = f"S3 path {source_path} does not exist as object or folder"
            if self.config.fail_if_missing:
                raise FileNotFoundError(message)
            self.logger.warning(message)
            if self.config.include_detailed_response:
                return DataSyncResult(bytes_transferred=0, files_transferred=0)
            else:
                return DataSyncResult()

        _sync_paths = sync_paths

        if self.config.require_lock:
            delay = 5
            tries = MAX_LOCK_WAIT_TIME_IN_SECS // delay
            self.logger.info(
                f"File lock required for transfer. Will attempt to aquire lock {tries} times, "
                f"with {delay} sec delays between attempts. "
            )

            @retry(CannotAcquirePathLockError, tries=tries, delay=delay, backoff=1)
            @functools.wraps(sync_paths)
            def sync_paths_with_lock(*args, **kwargs):
                with PathLock(destination_path, lock_root=os.getenv(LOCK_ROOT_ENV_VAR)):
                    response = sync_paths(*args, **kwargs)
                return response

            _sync_paths = sync_paths_with_lock

        remote_to_local_config = self.config.remote_to_local_config
        if source_is_object and remote_to_local_config.use_custom_tmp_dir:
            # If our source is an s3 object (not prefix) and we want to use custom object
            # download logic (default False), then we save s3 objects to a temporary location
            # that is on the SAME file system.
            #
            # This is necessary because if the normal boto3 download gets interrupted in a
            # catastrophic way that prevents built-in cleanup strategies, it leaves
            # a 'partial' file (e.g. `*.6eF5b5da`) that resides in the SAME parent directory
            # as the actual intended destination path. This 'partial' file can be picked up by
            # some scientific executables (e.g. cellranger) and interpreted as an invalid input
            if remote_to_local_config.custom_tmp_dir is None:
                custom_tmp_dir = find_filesystem_boundary(destination_path)
            elif isinstance(remote_to_local_config.custom_tmp_dir, EFSPath):
                custom_tmp_dir = self.sanitize_local_path(remote_to_local_config.custom_tmp_dir)
            else:
                custom_tmp_dir = remote_to_local_config.custom_tmp_dir

            with tempfile.TemporaryDirectory(dir=custom_tmp_dir) as tmp_dir:
                tmp_destination_path = Path(tmp_dir) / destination_path.name
                _sync_paths(
                    source_path=source_path,
                    destination_path=tmp_destination_path,
                    transfer_config=self.s3_transfer_config,
                    config=self.botocore_config,
                    force=self.config.force,
                    size_only=self.config.size_only,
                )
                destination_path.parent.mkdir(parents=True, exist_ok=True)
                os.rename(src=tmp_destination_path, dst=destination_path)
        else:
            # If our source is a prefix, then _sync_paths has builtin logic to deal with deleting
            # excess files in the destination dir that do not match the source prefix layout.
            _sync_paths(
                source_path=source_path,
                destination_path=destination_path,
                transfer_config=self.s3_transfer_config,
                config=self.botocore_config,
                force=self.config.force,
                size_only=self.config.size_only,
                delete=True,
            )

        self.logger.info(f"Updating last modified time on local files to at least {start_time}")
        refresh_local_path__mtime(destination_path, start_time.timestamp())

        if not self.config.retain_source_data:
            # TODO: maybe tag for deletion
            self.logger.warning(
                "Deleting s3 objects not allowed when downloading them to local file system"
            )

        result = DataSyncResult()
        # Collecting stats for detailed response
        if self.config.include_detailed_response:
            result.files_transferred = len(find_all_paths(destination_path, include_dirs=False))
            result.bytes_transferred = get_path_size_bytes(destination_path)
        return result

    def sync_local_to_local(
        self, source_path: LocalPath, destination_path: LocalPath
    ) -> DataSyncResult:
        source_path = self.sanitize_local_path(source_path)
        destination_path = self.sanitize_local_path(destination_path)
        self.logger.info(f"Copying local content from {source_path} -> {destination_path}")
        start_time = datetime.now(tz=timezone.utc)

        if not source_path.exists():
            if self.config.fail_if_missing:
                raise FileNotFoundError(f"Local path {source_path} does not exist")
            self.logger.warning(f"Local path {source_path} does not exist")
            return DataSyncResult(bytes_transferred=0)

        if self.config.retain_source_data:
            copy_path(source_path=source_path, destination_path=destination_path, exists_ok=True)
        else:
            move_path(source_path=source_path, destination_path=destination_path, exists_ok=True)
        self.logger.info(f"Updating last modified time on local files to at least {start_time}")
        refresh_local_path__mtime(destination_path, start_time.timestamp())

        result = DataSyncResult()
        # Collecting stats for detailed response
        if self.config.include_detailed_response:
            result.files_transferred = len(find_all_paths(source_path, include_dirs=False))
            result.bytes_transferred = get_path_size_bytes(source_path)
        return result

    def sync_s3_to_s3(
        self,
        source_path: S3URI,
        destination_path: S3URI,
        source_path_prefix: Optional[S3KeyPrefix] = None,
    ) -> DataSyncResult:
        self.logger.info(f"Syncing s3 content from {source_path} -> {destination_path}")

        if not is_object(source_path) and not is_folder(source_path):
            message = f"S3 path {source_path} does not exist as object or folder"
            if self.config.fail_if_missing:
                raise FileNotFoundError(message)
            self.logger.warning(message)
            if self.config.include_detailed_response:
                return DataSyncResult(bytes_transferred=0, files_transferred=0)
            else:
                return DataSyncResult()

        sync_paths(
            source_path=source_path,
            destination_path=destination_path,
            source_path_prefix=source_path_prefix,
            transfer_config=self.s3_transfer_config,
            config=self.botocore_config,
            force=self.config.force,
            size_only=self.config.size_only,
            delete=True,
        )
        if not self.config.retain_source_data:
            delete_s3_path(s3_path=source_path)

        result = DataSyncResult()
        if self.config.include_detailed_response:
            path_stats = get_s3_path_stats(destination_path)
            result.files_transferred = path_stats.object_count or 0
            result.bytes_transferred = path_stats.size_bytes
        return result

    def sync(
        self,
        source_path: Union[LocalPath, S3URI],
        destination_path: Union[LocalPath, S3URI],
        source_path_prefix: Optional[str] = None,
    ) -> DataSyncResult:
        if isinstance(source_path, S3URI) and isinstance(destination_path, S3URI):
            return self.sync_s3_to_s3(
                source_path=source_path,
                destination_path=destination_path,
                source_path_prefix=S3KeyPrefix(source_path_prefix) if source_path_prefix else None,
            )

        elif isinstance(source_path, S3URI):
            return self.sync_s3_to_local(
                source_path=source_path,
                destination_path=cast(LocalPath, destination_path),
            )
        elif isinstance(destination_path, S3URI):
            return self.sync_local_to_s3(
                source_path=cast(LocalPath, source_path),
                destination_path=destination_path,
            )
        else:
            return self.sync_local_to_local(
                source_path=source_path,
                destination_path=destination_path,
            )

    def sync_task(self, task: DataSyncTask) -> DataSyncResult:
        return self.sync(
            source_path=task.source_path,
            destination_path=task.destination_path,
            source_path_prefix=task.source_path_prefix,
        )

    @classmethod
    def sync_request(cls, request: DataSyncRequest) -> DataSyncResult:
        sync_operations = cls(config=request.config)
        return sync_operations.sync_task(task=request.task)

    # -----------------------------------
    # Helper methods
    # -----------------------------------

    def sanitize_local_path(self, path: Union[EFSPath, Path]) -> Path:
        if isinstance(path, EFSPath):
            self.logger.info(f"Sanitizing efs path {path}")
            new_path = get_local_path(path, raise_if_unmounted=True)
            self.logger.info(f"Sanitized efs path -> {new_path}")
            return new_path
        return path


# We should consider using cloudpathlib[s3] in the future
def sync_data(
    source_path: Union[S3URI, LocalPath],
    destination_path: Union[S3URI, LocalPath],
    source_path_prefix: Optional[str] = None,
    max_concurrency: int = 10,
    retain_source_data: bool = True,
    require_lock: bool = False,
    force: bool = False,
    size_only: bool = False,
    fail_if_missing: bool = True,
    remote_to_local_config: Optional[RemoteToLocalConfig] = None,
    include_detailed_response: bool = False,
):
    request = DataSyncRequest(
        source_path=source_path,
        destination_path=destination_path,
        source_path_prefix=S3KeyPrefix(source_path_prefix) if source_path_prefix else None,
        max_concurrency=max_concurrency,
        retain_source_data=retain_source_data,
        require_lock=require_lock,
        force=force,
        size_only=size_only,
        fail_if_missing=fail_if_missing,
        remote_to_local_config=remote_to_local_config or RemoteToLocalConfig(),
        include_detailed_response=include_detailed_response,
    )
    return DataSyncOperations.sync_request(request=request)


def refresh_local_path__mtime(path: Path, min_mtime: Union[int, float]):
    paths = find_all_paths(path, include_dirs=False, include_files=True)
    for subpath in paths:
        path_stats = os.stat(subpath)
        if path_stats.st_mtime < min_mtime:
            os.utime(subpath, times=(path_stats.st_atime, min_mtime))



S3_LANDING_BUCKET = "landing"
S3_ASSETS_BUCKET = "assets"
S3_FILE_STORE_BUCKET = "file-store"


S3_SOURCE_PATH_VAR = "S3_SOURCE_PATH"
S3_SOURCE_PATH_PREFIX_VAR = "S3_SOURCE_PATH_PREFIX"
S3_DESTINATION_PATH_VAR = "S3_DESTINATION_PATH"


# scratch file constants

S3_SCRATCH_KEY_PREFIX = "scratch/"
S3_SCRATCH_TAGGING_KEY = "time-to-live"
S3_SCRATCH_TAGGING_VALUE = "scratch"



from typing import TYPE_CHECKING, List, Optional

from aibs_informatics_aws_utils.core import AWSService, get_region
from aibs_informatics_aws_utils.exceptions import ResourceNotFoundError

if TYPE_CHECKING:  # pragma: no cover
    from mypy_boto3_apigateway.type_defs import RestApiTypeDef
else:
    RestApiTypeDef = dict

get_apigateway_client = AWSService.API_GATEWAY.get_client


def get_rest_api(api_name: str, region: Optional[str] = None) -> RestApiTypeDef:
    apigw = get_apigateway_client(region=region)

    paginator = apigw.get_paginator("get_rest_apis")
    rest_apis: List[RestApiTypeDef] = paginator.paginate(
        PaginationConfig={"MaxItems": 100}
    ).build_full_result()["items"]

    for rest_api in rest_apis:
        # In theory, only one api should be associated with env-base
        if rest_api.get("name") == api_name:
            return rest_api
    else:
        raise ResourceNotFoundError(f"Could not resolve REST Api with {api_name}")


def get_rest_api_endpoint(
    rest_api: RestApiTypeDef, stage: str = "prod", region: Optional[str] = None
) -> str:
    api_id = rest_api["id"]  # type: ignore  # mypy_boto3 TypeDict makes optional, but actually is required
    region = get_region(region)
    return f"https://{api_id}.execute-api.{region}.amazonaws.com/{stage}"



from __future__ import annotations

import re
from math import ceil
from re import Pattern

__all__ = [
    "get_file_system",
    "list_file_systems",
]
import logging
from pathlib import Path
from typing import TYPE_CHECKING, ClassVar, Dict, List, Optional, Tuple, Union

from aibs_informatics_core.collections import ValidatedStr

from aibs_informatics_aws_utils.core import AWSService

if TYPE_CHECKING:  # pragma: no cover
    from mypy_boto3_fsx.type_defs import (
        DataRepositoryAssociationTypeDef,
        FileSystemTypeDef,
        FilterTypeDef,
        TagTypeDef,
    )
else:
    FileSystemTypeDef = dict
    TagTypeDef = dict


logger = logging.getLogger(__name__)

get_fsx_client = AWSService.FSX.get_client

StrPath = Union[Path, str]


class FSxFileSystemId(ValidatedStr):
    regex_pattern: ClassVar[Pattern] = re.compile(r"fs-[0-9a-f]{17}")


FileSystemIdOrName = Union[FSxFileSystemId, str]
FileSystemNameOrId = Union[FSxFileSystemId, str]


def split_name_and_id(
    id_or_name: Optional[FileSystemNameOrId],
) -> Tuple[Optional[str], Optional[FSxFileSystemId]]:
    """Identify file system identifier as name or id

    Examples:
        INP: NAME1
        OUT: (NAME1, None)

        INP: ID1
        OUT: (None, ID1)

        INP: None
        OUT: (None, None)

    Args:
        id_or_name (Optional[FileSystemNameOrId]): File system id or name.


    Returns:
        Tuple[Optional[str], Optional[FSxFileSystemId]]: Tuple of name and id.
    """
    if not id_or_name:
        return None, None
    elif FSxFileSystemId.is_valid(id_or_name):
        return None, FSxFileSystemId(id_or_name)
    return id_or_name, None


def split_name_and_ids(
    names_or_ids: List[FileSystemNameOrId],
) -> Tuple[List[str], List[FSxFileSystemId]]:
    """Split file system combined list of names and ids into separate lists.

    Example:
        INP: [NAME1, ID1, NAME2, NAME3, ID2]
        OUT: ([NAME1, NAME2, NAME3], [ID1, ID2])

    Args:
        names_or_ids (List[FileSystemNameOrId]): List of names and/or ids.

    Returns:
        Tuple[List[str], List[FSxFileSystemId]]: Tuple of name and id lists.
    """
    if not names_or_ids:
        return [], []
    names, file_system_ids = zip(*(split_name_and_id(name_or_id) for name_or_id in names_or_ids))
    return [name for name in names if name], [
        file_system_id for file_system_id in file_system_ids if file_system_id
    ]


def resolve_file_system_ids(*name_or_ids: FileSystemNameOrId) -> List[FSxFileSystemId]:
    """Resolve file system ids from file system names or ids.

    Args:
        name_or_ids (Tuple[FileSystemNameOrId]): File system names or ids.

    Returns:
        str: File system id.
    """
    file_system_ids: List[FSxFileSystemId] = []
    for name_or_id in name_or_ids:
        name, file_system_id = split_name_and_id(name_or_id)
        if file_system_id:
            file_system_ids.append(file_system_id)
        else:
            file_system = get_file_system(name)
            if not file_system:
                raise ValueError(f"File system not found with name: {name}")
            file_system_ids.append(FSxFileSystemId(file_system.get("FileSystemId", "")))
    return file_system_ids


def get_file_system(
    name_or_id: Optional[FileSystemNameOrId] = None,
    tags: Optional[Dict[str, str]] = None,
) -> FileSystemTypeDef:
    """Get FSx file system.

    Args:
        file_system_id (Optional[str], optional): File system id.
        name (Optional[str], optional): File system name.
        tags (Optional[Dict[str, str]], optional): File system tags.

    Returns:
        FileSystemDescriptionTypeDef: File system description.
    """
    if not name_or_id and not tags:
        raise ValueError("At least one of file_system_id, name or tags must be provided.")

    file_systems = list_file_systems(name_or_ids=[name_or_id] if name_or_id else None, tags=tags)
    if len(file_systems) > 1:
        raise ValueError(
            f"Multiple file systems found with name/id: {name_or_id} and tags: {tags}"
        )
    if not file_systems:
        raise ValueError(f"File system not found with name/id: {name_or_id} and tags: {tags}")
    return file_systems[0]


def list_file_systems(
    name_or_ids: Optional[List[FileSystemNameOrId]] = None,
    tags: Optional[Dict[str, str]] = None,
    **kwargs,
) -> List[FileSystemTypeDef]:
    """List FSx file systems.

    You can filter on id, name and tags.

    Args:
        file_system_id (Optional[str], optional): Optionally filter on file system id.
        name (Optional[str], optional): Optionally filter on name.
        tags (Optional[Dict[str, str]], optional): Optionally filter on tags.

    Returns:
        List[FileSystemDescriptionTypeDef]: List of file systems.
    """
    client = get_fsx_client(**kwargs)
    paginator = client.get_paginator("describe_file_systems")
    names, file_system_ids = split_name_and_ids(name_or_ids or [])
    if file_system_ids:
        response_iter = paginator.paginate(FileSystemIds=file_system_ids)
    else:
        response_iter = paginator.paginate()
    file_systems: List[FileSystemTypeDef] = []
    for response in response_iter:
        filtered_file_systems = response["FileSystems"]
        if not names and not tags:
            file_systems.extend(filtered_file_systems)
            continue
        old_filtered_file_systems = filtered_file_systems
        filtered_file_systems = []
        for fs in old_filtered_file_systems:
            fs_tags_dict = {tag["Key"]: tag["Value"] for tag in fs.get("Tags", [])}
            if names and fs_tags_dict.get("Name") not in names:
                continue
            if tags and not all(
                tag in fs_tags_dict and value == fs_tags_dict[tag] for tag, value in tags.items()
            ):
                continue
            filtered_file_systems.append(fs)
        file_systems.extend(filtered_file_systems)
    return file_systems


def list_data_repository_associations(
    name_or_id: Optional[FileSystemIdOrName],
    filters: Optional[List[FilterTypeDef]] = None,
    # TODO: should I include data repository paths?
    data_repository_paths: Optional[List[str]] = None,
    **kwargs,
) -> List[DataRepositoryAssociationTypeDef]:
    """List data repository associations for a file system.

    Args:
        file_system_id (str): File system id.

    Returns:
        List[FileSystemTypeDef]: List of data repository associations.
    """
    client = get_fsx_client(**kwargs)
    if name_or_id:
        file_system_id = resolve_file_system_ids(name_or_id)
        if filters:
            if id_filter := next(
                (filter for filter in filters if filter.get("Name") == "file-system-id"), None
            ):
                id_filter["Values"] = list(id_filter.get("Values", [])) + [str(file_system_id)]
            else:
                filters.append({"Name": "file-system-id", "Values": file_system_id})
        else:
            filters = [{"Name": "file-system-id", "Values": file_system_id}]
    if not filters:
        filters = []
    associations: List[DataRepositoryAssociationTypeDef] = []
    response = client.describe_data_repository_associations(Filters=filters)
    while response["Associations"]:
        new_associations = response["Associations"]
        if data_repository_paths:
            new_associations = [
                association
                for association in new_associations
                if association.get("DataRepositoryPath") in data_repository_paths
            ]
        associations.extend(new_associations)
        if not (next_token := response.get("NextToken")):
            break
        response = client.describe_data_repository_associations(
            Filters=filters, NextToken=next_token
        )
    return associations


def calculate_size_required(bytes_required: int) -> int:
    """Calculate size of file system for the given bytes specified.

    FSx file systems are created with a size of
        - 1.2 TB,
        - 2.4 TB
        - any multiple 2.4 TB.

    Args:
        bytes_required (int): Bytes required.

    Returns:
        int: Size required.
    """
    BYTES_IN_TB = 1024 * 1024 * 1024 * 1024
    if bytes_required <= 1.2 * BYTES_IN_TB:
        return ceil(1.2 * BYTES_IN_TB)
    if bytes_required <= 2.4 * BYTES_IN_TB:
        return ceil(2.4 * BYTES_IN_TB)
    return ceil((bytes_required // (2.4 * BYTES_IN_TB) + 1) * 2.4 * BYTES_IN_TB)
